package org.netkernel.xml.accessor;

// Created by IntelliJ IDEA.
// Author: Tom Geudens
// Date: 2011/11/17

// The usual suspects for an accessor.
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.meta.impl.SourcedArgumentMetaImpl;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.xml.util.XMLUtils;

// HDS in action.
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;

// FastXPath Implementation.
import org.netkernel.layer0.util.FastXPath;

// For the operator argument.
import org.netkernel.layer1.representation.IDeterminateStringRepresentation;

// Xalan XPath Implementation.
import org.apache.xpath.*;
import org.apache.xpath.objects.*;

// It doesn't get DOMmer than this.
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

// For completeness.
import java.lang.String;
import java.lang.Boolean;
import java.util.List;

public class XPath1 extends StandardAccessorImpl
{
	// Two mandatory arguments, operator and operand
	// Three possible representations, document, boolean or string,
	public XPath1()
	{
		this.declareThreadSafe();
		this.declareSourceRepresentation(Document.class);
		this.declareSourceRepresentation(Boolean.class);
		this.declareSourceRepresentation(String.class);
		declareArgument(new SourcedArgumentMetaImpl("operator", null, null, new Class[] { String.class }));
		declareArgument(new SourcedArgumentMetaImpl("operand", null, null, new Class[] { Document.class }));
	}

	// Both Xalan and FastXPath return a NodeList.
	// But not exactly the same NodeList.
	// This small private class (borrowed from DOMXDA code)
	// addresses that problem.
	private class NodeListImpl implements NodeList
	{
		private List mList;

		public NodeListImpl(List aList)
		{
			mList = aList;
		}

		public int getLength()
		{
			return mList.size();
		}

		public Node item(int index)
		{
			return (Node) mList.get(index);
		}
	}

	// The above class addresses the problem, this method benefits
	// from it. All XPath Queries returning a Document pass
	// through here.
	private Document parseNodeList(NodeList cResult, INKFRequestContext lContext) throws Exception
	{
		Document d = XMLUtils.newDocument();
		Element docEl=d.createElement("resultset");
		d.appendChild(docEl);
		for (int i = 0; i < cResult.getLength(); i++)
		{	Node thisNode = cResult.item(i);
			docEl.appendChild(d.importNode(thisNode,true));
		}
		return d;
	};

	// Only query, only SOURCE
	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		INKFRequestReadOnly thisReq = aContext.getThisRequest();

		// Three possible Representations.
		String sResult = null;
		Document dResult = null;
		Boolean bResult = null;

		// Two mandatory Arguments.
		String aXPath = aContext.source("arg:operator", IDeterminateStringRepresentation.class).getString();
		Document aDocument = aContext.source("arg:operand", Document.class);

		// One response.
		INKFResponse response;

		// Determine the processing
		if (FastXPath.isSuitable(aXPath))
		{
			// FastXPath from Layer0 in action.
			NodeList cResult;
			cResult = new NodeListImpl(FastXPath.eval(aDocument, aXPath));
			dResult = parseNodeList(cResult, aContext);
		}
		else
		{
			// Xalan XPath from Apache in action.
			XObject x;
			XNodeSetForDOM xResult;
			NodeList cResult;

			x = XPathAPI.eval(aDocument, aXPath);

			if (x.getType() == XObject.CLASS_BOOLEAN)
			{
				bResult = (Boolean) x.castToType(XObject.CLASS_BOOLEAN, null);
			}
			else if (x.getType() == XObject.CLASS_STRING)
			{
				sResult = (String) x.castToType(XObject.CLASS_STRING, null);
			}
			else if ((x.getType() == XObject.CLASS_NODESET) || (x.getType() == XObject.CLASS_NULL))
			{
				xResult = new XNodeSetForDOM((XNodeSet) x.castToType(XObject.CLASS_NODESET, null));
				cResult = xResult.nodelist();
				dResult = parseNodeList(cResult, aContext);
			}
		}

		// Determine Representation
		if (!(dResult == null))
		{
			// For testing purposes :
			// String subrequest_url = "meta:apache.XMLTidy";
			// INKFRequest subrequest = aContext.createRequest(subrequest_url);
			// subrequest.setVerb(INKFRequestReadOnly.VERB_SOURCE);
			// subrequest.addArgumentByValue("operand",dResult);
			// subrequest.setRepresentationClass(Document.class);
			// dResult = (Document)aContext.issueRequest(subrequest);
			response = aContext.createResponseFrom(dResult);
		}
		else if (!(bResult == null))
		{
			response = aContext.createResponseFrom(bResult);
		}
		else if (!(sResult == null))
		{
			response = aContext.createResponseFrom(sResult);
		}
	}
}
